/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2005 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * AbstractInfoScreen.java
 *
 * Created on August 26, 2005, 1:53 PM
 *
 */

package org.netbeans.microedition.lcdui;

import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

/**
 * An abstract class serving as a parent for <code>SplashScreen</code>
 * and <code>WaitScreen</code>. This class provides the basic visualization
 * of the screen and can perform automatic switch to a next displayable when
 * the screen is dismissed (either programatically or by user action).
 * <p/>
 * When this screen is displayed, it can display either supplied text or image.
 * The current implementation shows both, text and image, centered
 * in the middle of the display.
 * @author breh
 */
public abstract class AbstractInfoScreen extends Canvas {
	
	private Display display;
	
	private Image image;	
	private String text;
	
	private Font DEFAULT_TEXT_FONT = Font.getFont(Font.FONT_STATIC_TEXT);
	private Font textFont = DEFAULT_TEXT_FONT;
	
	private Display previouslyShownDisplayable;
	private Displayable nextDisplayable;
	private Alert nextAlert;	
	
	private Displayable previousDisplayable;
	
	/**
	 * Creates a new instance of AbstractInfoScreen
	 * @param display display parameter. Cannot be null
	 * @throws java.lang.IllegalArgumentException if the display parameter is null
	 */
	public AbstractInfoScreen(Display display) throws IllegalArgumentException  {
		if (display == null) throw new IllegalArgumentException("Display parameter cannot be null."); // NOI18N
		this.display = display;
	}
	
	
	// properties
	
	/**
	 * Sets the text to be painted on the screen.
	 * @param text text to be painter, or null if no text should be shown
	 */
	public void setText(String text) {
		this.text = text;
		repaint();
	}
	
	/**
	 * Gets the text to be painted on the screen.
	 * @return text
	 */
	public String getText() {
		return this.text;
	}
	
	
	/**
	 * Gets the image to be painted on the screen.
	 * @return image
	 */
	public Image getImage() {
		return image;
	}
	
	/**
	 * Sets image to be painted on the screen. If set to null, no image
	 * will be painted
	 * @param image image to be painted. Can be null.
	 */
	public void setImage(Image image) {
		this.image = image;
		repaint();
	}
	
	/**
	 * Sets the font to be used to paint the specified text. If set 
	 * to null, the default font (Font.STATIC_TEXT_FONT) will be used.
	 * @param font font to be used to paint the text. May be null.
	 */
	public void setTextFont(Font font) {
		if (font != null) {
			this.textFont = font;
		} else {
			this.textFont = DEFAULT_TEXT_FONT;
		}
		repaint();
	}
	
	/**
	 * Gets the current font used to paint the text.
	 * @return text font
	 */
	public Font getTextFont() {
		return textFont;
	}
	
	/**
	  * Sets the displayable to be used to switch when the screen is being dismissed.
	 * @param nextDisplayable displayable, or null if the component should switch back
	 * to the screen from which was displayed prior showing this component.
	 */
	public void setNextDisplayable(Displayable nextDisplayable) {
		this.nextDisplayable = nextDisplayable;
	}
	
	
	/**
	 * Requests that the specified Alert is going to be shown in the case of
	 * screen dismiss, and nextDisplayable be made current after the Alert is dismissed.
	 *  <p/>
	 * The nextDisplayable parameter cannot be Alert and in the case
	 * nextAlert is not null, it also cannot be null.
	 * @param nextAlert alert to be shown, or null if the component should return back to the original screen
	 * @param nextDisplayable a displayable to be shown after the alert is being dismissed. This displayable
	 * cannot be null if the <code>nextAlert</code> is not null and it also cannot be 
	 * Alert.
	 * @throws java.lang.IllegalArgumentException If the nextAlert is not null and nextDisplayable is null at the same time, or
	 * if the nextDisplayable is instance of <code>Alert</code>
	 */
	public void setNextDisplayable(Alert nextAlert, Displayable nextDisplayable) throws IllegalArgumentException {
		if ((nextAlert != null) && (nextDisplayable == null))
			throw new IllegalArgumentException("A nextDisplayable parameter cannot be null if the nextAlert parameter is not null.");
		if (nextDisplayable instanceof Alert)
			throw new IllegalArgumentException("nextDisplayable paremter cannot be Alert.");
		this.nextAlert = nextAlert;
		this.nextDisplayable  = nextDisplayable;
	}	
	
	// protected methods

	/**
	 * implementation of abstract method
	 * @param g 
	 */
	protected void paint(Graphics g) {		
		g.setColor(display.getColor(Display.COLOR_BACKGROUND));
		g.fillRect(0,0,g.getClipWidth(), g.getClipHeight());
		g.setColor(display.getColor(Display. COLOR_FOREGROUND));
		final int centerX = g.getClipWidth() / 2 + g.getClipX();
		final int centerY = g.getClipHeight() / 2 + g.getClipY();
		if (image != null) {	
			g.drawImage(image, centerX, centerY, Graphics.HCENTER | Graphics.VCENTER);
		}
		if (text != null) {
			g.drawString(text, centerX, centerY, Graphics.HCENTER | Graphics.BASELINE);
		}
	}	
	
	/**
	 * repaints the screen whem a size has changed.
	 */
	protected void sizeChanged(int w, int h) {
		repaint();
	}	
	
	
	/**
	 * Gets the used display object
	 * @return display object
	 */
	protected Display getDisplay() {
		return display;
	}
	
	
	/**
	 * Gets the next displayable
	 * @return 
	 */
	protected Displayable getNextDisplayable() {
		return nextDisplayable;
	}
	
	
	/**
	 * gets the next alert
	 * @return 
	 */
	protected Alert getNextAlert() {
		return nextAlert;
	}
	
	
	/**
	 * switch to the next displayable (or alert)
	 */
	protected void switchToNextDisplayable() {
		if (nextDisplayable != null) {
			switchToDisplayable(display, nextAlert, nextDisplayable);
		} else if (previousDisplayable != null) {
			display.setCurrent(previousDisplayable);
		}		
	}
	
	/**
	 * Switch to the given displayable and alert
	 * @param display 
	 * @param alert 
	 * @param displayable 
	 */
	protected static void switchToDisplayable(Display display, Alert alert, Displayable displayable) {
		if (displayable != null) {
			if (alert != null) {
				display.setCurrent(alert,displayable);
			} else {
				display.setCurrent(displayable);
			}
		}
	}	

	/**
	 * sets value of previous displayable. Implementation should always 
	 * call this super implementation when overriding this method
	 */
	protected void showNotify() {
		previousDisplayable = getDisplay().getCurrent();
		super.showNotify();
	}
	
	
	
	
}
