/*
 *                 Sun Public License Notice
 *
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 *
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2005 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

/*
 * SplashScreen.java
 *
 * Created on August 26, 2005, 10:19 AM
 */

package org.netbeans.microedition.lcdui;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

/**
 * This component represents a splash screen, which is usually being displayed
 * when the application starts.
 * @author breh
 */
public class SplashScreen extends AbstractInfoScreen {

	private static final int DEFAULT_TIMEOUT = 5000;
	
	private int timeout = DEFAULT_TIMEOUT;
	private boolean allowTimeoutInterrupt = true;
	
	private long currentDisplayTimestamp;
		
	/**
	 * Creates a new instance of SplashScreen
	 * @param display display - cannot be null
	 * @throws java.lang.IllegalArgumentException when the display parameter is null
	 */
	public SplashScreen(Display display) throws IllegalArgumentException  {
		super(display);
	}

	// properties
		
	
	/**
	 * Sets the timeout of the splash screen - i.e. the time in milliseconds for
	 * how long the splash screen is going to be shown on the display.
	 * <p/>
	 * If the supplied timeout is 0, then the splashscreen waits forever (it needs to
	 * be dismissed by pressing a key)
	 * @param timeout timeout in milliseconds
	 */
	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}
	
	/**
	 * Gets current timeout of the splash screen
	 * @return timeout value
	 */
	public int getTimeout() {
		return timeout;
	}	
	
	/**
	 * Set to true, when the timeout with a specified timeout interval can
	 * be interrupted by pressing a key.
	 * @param allow true if the user can interrupt the screen, false if the user need to wait
	 * until timeout.
	 */
	public void setAllowTimeoutInterrupt(boolean allow) {
		this.allowTimeoutInterrupt = allow;
	}
	
	/**
	 * Can be the splashscreen interrupted (dismissed) by the user pressing a key?
	 * @return true if user can interrupt it, false otherwise
	 */
	public boolean isAllowTimeoutInterrupt() {
		return allowTimeoutInterrupt;
	}
	
	
	// canvas methods
	
	/**
	 * keyPressed callback
	 * @param keyCode 
	 */
	protected void keyPressed(int keyCode) {
		//
		userAction();
	}
	

	/**
	 * pointerPressed callback
	 * @param x 
	 * @param y 
	 */
	protected void pointerPressed(int x, int y) {
		//
		userAction();
	}	

	/**
	 * starts the coundown of the timeout
	 */	
	protected void showNotify() {
		super.showNotify();
		// start watchdog task - only when applicable
		currentDisplayTimestamp = System.currentTimeMillis();		
		if (timeout > 0) {
			Watchdog w = new Watchdog(timeout, currentDisplayTimestamp);
			w.start();
		}
	}
	

	protected void hideNotify() {
		super.hideNotify();
		currentDisplayTimestamp = System.currentTimeMillis();
	}
	
	

	// private stuff
		
	private void userAction() {
		if (allowTimeoutInterrupt) {
			switchToNextDisplayable();
		}
	}
	
	
	private void timeout() {
		switchToNextDisplayable();
	}


	
	private class Watchdog extends Thread {
		
		private int timeout;
		private long currentDisplayTimestamp;
		
		private Watchdog(int timeout, long currentDisplayTimestamp) {
			this.timeout = timeout;
			this.currentDisplayTimestamp = currentDisplayTimestamp;
		}
		
		public void run() {
			try {
				Thread.sleep(timeout);
			} catch (InterruptedException ie) {
			}
			// timeout (only if current display timout matches) - this means this
			// splash screen is still being shown on the display
			if (this.currentDisplayTimestamp == SplashScreen.this.currentDisplayTimestamp) {
				timeout();
			}
		}
		
		
	}
	
	
}
